#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "db_edit.h"
#include "db_compiler.h"
#include "db_vm.h"
#include "db_buffer.h"

#define FILEIO

/* command handlers */
static void DoNew(System *sys);
static void DoLoad(System *sys);
static void DoSave(System *sys);
static void DoList(System *sys);
static void DoRun(System *sys);
static void DoQuit(System *sys);

typedef void CmdHandlerFcn(System *sys);
#define CMD_NULL    ((CmdHandlerFcn *)0)

/* command table */
static struct {
    char *name;
    CmdHandlerFcn *handler;
} cmds[] = {
    {   "NEW",      DoNew       },
    {   "LOAD",     DoLoad      },
    {   "SAVE",     DoSave      },
    {   "LIST",     DoList      },
    {   "RUN",      DoRun       },
    {   "QUIT",     DoQuit      },
    {   NULL,       CMD_NULL    }
};

/* prototypes */
static void EditRewind(void *cookie);
static int EditGetLine(void *cookie, char *buf, int len, int *pLineNumber);
static char *NextToken(System *sys);
static int ParseNumber(System *sys, char *token, int *pValue);
static int IsBlank(char *p);

void EditWorkspace(System *sys)
{
    int lineNumber;
    char *token;

    VM_printf("xBasic 0.001\n");

    for (;; ) {
        
        VM_getline(sys->lineBuf, sizeof(sys->lineBuf));

        sys->linePtr = sys->lineBuf;

        if ((token = NextToken(sys)) != NULL) {
            if (ParseNumber(sys, token, &lineNumber)) {
                if (IsBlank(sys->linePtr)) {
                    if (!EditBufDeleteLineN(lineNumber))
                        VM_printf("no line %d\n", lineNumber);
                }
                else if (!EditBufAddLineN(lineNumber, sys->linePtr))
                    VM_printf("out of edit buffer space\n");
            }

            else {
                int i;
                for (i = 0; cmds[i].name != NULL; ++i)
                    if (strcasecmp(token, cmds[i].name) == 0)
                        break;
                if (cmds[i].handler) {
                    (*cmds[i].handler)(sys);
                    VM_printf("OK\n");
                }
                else
                    VM_printf("unknown command: %s\n", token);
            }
        }
    }
}

static void DoNew(System *sys)
{
    EditBufReset();
}

static void DoLoad(System *sys)
{
    char *name;
    if (!(name = NextToken(sys)))
        VM_printf("Please specify a file to load\n");
    else {
#ifdef FILEIO
        FILE *fp;
        if (!(fp = fopen(name, "r")))
            VM_printf("error loading '%s'\n", name);
        else {
            VM_printf("Loading '%s'\n", name);
            EditBufReset();
            while (fgets(sys->lineBuf, sizeof(sys->lineBuf), fp) != NULL) {
                int lineNumber;
                char *token;
                sys->linePtr = sys->lineBuf;
                if ((token = NextToken(sys)) != NULL) {
                    if (ParseNumber(sys, token, &lineNumber))
                        EditBufAddLineN(lineNumber, sys->linePtr);
                    else
                        VM_printf("expecting a line number: %s\n", token);
                }
            }
            fclose(fp);
        }
#endif
    }
}

static void DoSave(System *sys)
{
    char *name;
    if (!(name = NextToken(sys)))
        VM_printf("Please specify a file to save\n");
    else {
#ifdef FILEIO
        FILE *fp;
        if (!(fp = fopen(name, "w")))
            VM_printf("error creating '%s'\n", name);
        else {
            int lineNumber;
            VM_printf("Saving '%s'\n", name);
            EditBufSeekN(0);
            while (EditBufGetLine(&lineNumber, sys->lineBuf))
                fprintf(fp, "%d %s", lineNumber, sys->lineBuf);
            fclose(fp);
        }
#endif
    }
}

static void DoList(System *sys)
{
    int lineNumber;
    EditBufSeekN(0);
    while (EditBufGetLine(&lineNumber, sys->lineBuf))
        VM_printf("%d %s", lineNumber, sys->lineBuf);
}

static void DoRun(System *sys)
{
    ParseContext *c;
    if (!(c = InitCompiler(sys)))
        VM_printf("insufficient memory\n");
    else {
        ImageHdr *image;
        c->rewind = EditRewind;
        c->getLine = EditGetLine;
        c->getLineCookie = NULL;
        if ((image = Compile(c, MAX_OBJECTS)) != NULL) {
            size_t freeSize, stackSize;
            Interpreter *i;
            i = (Interpreter *)AllocateAllFreeSpace(sys, &freeSize);
            stackSize = (freeSize - sizeof(Interpreter)) / sizeof(VMVALUE);
            if (stackSize <= 0)
                VM_printf("insufficient memory\n");
            else {
                InitInterpreter(i, stackSize);
                Execute(i, image);
            }
        }
    }
    ResetFreeSpace(sys);
}

static void DoQuit(System *sys)
{
    exit(0);
}

static void EditRewind(void *cookie)
{
    EditBufSeekN(0);
}

static int EditGetLine(void *cookie, char *buf, int len, int *pLineNumber)
{
    return EditBufGetLine(pLineNumber, buf);
}

static char *NextToken(System *sys)
{
    char *token;
    int ch;
    while ((ch = *sys->linePtr) != '\0' && isspace(ch))
        ++sys->linePtr;
    token = sys->linePtr;
    while ((ch = *sys->linePtr) != '\0' && !isspace(ch))
        ++sys->linePtr;
    if (*sys->linePtr != '\0')
        *sys->linePtr++ = '\0';
    return *token == '\0' ? NULL : token;
}

static int ParseNumber(System *sys, char *token, int *pValue)
{
    int ch;
    *pValue = 0;
    while ((ch = *token++) != '\0' && isdigit(ch))
        *pValue = *pValue * 10 + ch - '0';
    return ch == '\0';
}

static int IsBlank(char *p)
{
    int ch;
    while ((ch = *p++) != '\0')
        if (!isspace(ch))
            return VMFALSE;
    return VMTRUE;
}


