/* db_image.h - compiled image file definitions
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#ifndef __DB_IMAGE_H__
#define __DB_IMAGE_H__

#include <stdarg.h>
#include "db_system.h"
#include "db_types.h"

/* stack frame offsets */
#define F_CODE  1
#define F_FP    2
#define F_PC    3
#define F_SIZE  3

/* intrinsic function flag */
#define INTRINSIC_FLAG  (1 << (sizeof(VMVALUE) * 8 - 1))

/* primitive prototypes */
#define PROTO_CODE      -1
#define PROTO_VECTOR    -2
#define PROTO_BVECTOR   -3

/* object header */
typedef struct ObjectHdr ObjectHdr;

/* object handle structure */
typedef struct {
    ObjectHdr *hdr;
} ObjectHandle;

/* object header structure */
struct ObjectHdr {
    ObjectHandle *handle;
    VMVALUE prototype;
    VMUVALUE size;
};

/* in-memory image header */
typedef struct {
    VMVALUE         mainCode;       /* main code object number */
    VMVALUE         *variables;     /* variable table */
    VMUVALUE        variableCount;  /* variable count */
    VMUVALUE        nHandles;       /* heap - number of handles */
    ObjectHandle    *handles;       /* heap - array of handles */
    ObjectHandle    *endHandles;    /* heap - end of the array of handles */
    ObjectHandle    *freeHandles;   /* heap - list of free handles */
    uint8_t         *data;          /* heap - heap data */
    uint8_t         *free;          /* heap - next free heap location */
} ImageHdr;

/* get an object handle */
#define GetObjHandle(i, n)      ((i)->endHandles - (n))

/* get a pointer to an object header */
#define GetObjHdr(i, n)         (GetObjHandle((i), (n))->hdr)

/* get the prototype of an object */
#define GetObjPrototype(h)      ((h)->prototype)

/* get the size of a vector */
#define GetVectorSize(h)        ((h)->size)

/* get the base of a word vector */
#define GetVectorBase(h)        ((VMVALUE *)((uint8_t *)(h) + sizeof(ObjectHdr)))

/* get the base of a byte vector */
#define GetBVectorBase(h)       ((uint8_t *)(h) + sizeof(ObjectHdr))

/* intrinsic function numbers */
enum {
    FN_ABS              = 0x00,
    FN_RND,
    FN_LEFT,
    FN_RIGHT,
    FN_MID,
    FN_CHR,
    FN_STR,
    FN_ASC,
    FN_LEN
};

/* register numbers */
enum {
    RG_CNT              = 0x00,
    RG_DIRA,
    RG_OUTA,
    RG_INA,
    RG_CLKFREQ
};

/* opcodes */
#define OP_HALT         0x00    /* halt */
#define OP_BRT          0x01    /* branch on true */
#define OP_BRTSC        0x02    /* branch on true (for short circuit booleans) */
#define OP_BRF          0x03    /* branch on false */
#define OP_BRFSC        0x04    /* branch on false (for short circuit booleans) */
#define OP_BR           0x05    /* branch unconditionally */
#define OP_NOT          0x06    /* logical negate top of stack */
#define OP_NEG          0x07    /* negate */
#define OP_ADD          0x08    /* add two numeric expressions */
#define OP_SUB          0x09    /* subtract two numeric expressions */
#define OP_MUL          0x0a    /* multiply two numeric expressions */
#define OP_DIV          0x0b    /* divide two numeric expressions */
#define OP_REM          0x0c    /* remainder of two numeric expressions */
#define OP_BNOT         0x0d    /* bitwise not of two numeric expressions */
#define OP_BAND         0x0e    /* bitwise and of two numeric expressions */
#define OP_BOR          0x0f    /* bitwise or of two numeric expressions */
#define OP_BXOR         0x10    /* bitwise exclusive or */
#define OP_SHL          0x11    /* shift left */
#define OP_SHR          0x12    /* shift right */
#define OP_LT           0x13    /* less than */
#define OP_LE           0x14    /* less than or equal to */
#define OP_EQ           0x15    /* equal to */
#define OP_NE           0x16    /* not equal to */
#define OP_GE           0x17    /* greater than or equal to */
#define OP_GT           0x18    /* greater than */
#define OP_LIT          0x19    /* load literal */
#define OP_GREF         0x1a    /* load a global variable */
#define OP_GSET         0x1b    /* set a global variable */
#define OP_LREF         0x1c    /* load a local variable relative to the frame pointer */
#define OP_LSET         0x1d    /* set a local variable relative to the frame pointer */
#define OP_VREF         0x1e    /* load an element of a vector */
#define OP_VSET         0x1f    /* set an element of a vector */
#define OP_RESERVE      0x20    /* reserve space on the stack */
#define OP_CALL         0x21    /* call a function */
#define OP_RETURN       0x22    /* return from a function leaving an integer result on the stack */
#define OP_DROP         0x23    /* drop the top element of the stack */
#define OP_GETREG       0x24    /* get register value */
#define OP_SETREG       0x25    /* set register value */
#define OP_STRING       0x26    /* string opcode prefix */
#define OP_PRINT        0x27    /* various print operations */

/* OP_PRINT operations */
enum {
    PRINT_INT = 0x00,
    PRINT_STR,
    PRINT_TAB,
    PRINT_NL
};

/* prototypes from db_vmheap.c */
int InitHeap(System *sys, ImageHdr *image, VMUVALUE size, VMUVALUE nHandles);
void ReserveHeapSpace(ImageHdr *image, VMUVALUE dataSize, VMUVALUE nHandles);
VMVALUE StringAlloc(ImageHdr *image, VMUVALUE length);
VMUVALUE GetObjectSize(ObjectHdr *hdr);
void CompactHeap(ImageHdr *image);
void DumpHeap(ImageHdr *image);

void VM_printf(const char *fmt, ...);           /* fmt in FLASH_SPACE */
void VM_vprintf(const char *fmt, va_list ap);   /* fmt in FLASH_SPACE */
void VM_putchar(int ch);

#endif
