/* db_vmint.c - bytecode interpreter for a simple virtual machine
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include "db_vm.h"
#include "db_vmdebug.h"

/* prototypes for local functions */
static void StartCode(Interpreter *i, VMVALUE object);
static void DoStringOp(Interpreter *i, int op);
static VMVALUE StringCat(Interpreter *i);
static int StringCmp(Interpreter *i);
static void DoPrint(Interpreter *i, int op);
static void PrintInteger(Interpreter *i, VMVALUE value);
static void PrintString(Interpreter *i, VMVALUE value);
static void PrintTab(Interpreter *i);
static void PrintC(Interpreter *i, int ch);

/* InitInterpreter - initialize the interpreter */
uint8_t *InitInterpreter(Interpreter *i, size_t stackSize)
{
    i->stack = (VMVALUE *)((uint8_t *)i + sizeof(Interpreter));
    i->stackTop = i->stack + stackSize;
    return (uint8_t *)i->stackTop;
}

/* Execute - execute the main code */
int Execute(Interpreter *i, ImageHdr *image)
{
    ObjectHdr *hdr;
    VMVALUE tmp, ind;
    int8_t tmpb;
    int cnt;

	/* setup the new image */
	i->image = image;

    /* initialize */
    i->code = i->image->mainCode;
    CheckObjNumber(i, i->code);
    hdr = GetObjHdr(i->image, i->code);
    CheckObjType(i, hdr, PROTO_CODE);
    i->cbase = i->pc = GetBVectorBase(hdr);
    i->sp = i->fp = i->stack - 1;
    i->linePos = 0;

    if (setjmp(i->errorTarget))
        return VMFALSE;

    for (;;) {
#if 0
        ShowStack(i);
        DecodeInstruction(i->code, i->cbase, i->pc);
#endif
        switch (VMCODEBYTE(i->pc++)) {
        case OP_HALT:
            return VMTRUE;
        case OP_BRT:
            for (tmp = 0, cnt = sizeof(VMUVALUE); --cnt >= 0; )
                tmp = (tmp << 8) | VMCODEBYTE(i->pc++);
            if (i->tos)
                i->pc += tmp;
            i->tos = Pop(i);
            break;
        case OP_BRTSC:
            for (tmp = 0, cnt = sizeof(VMUVALUE); --cnt >= 0; )
                tmp = (tmp << 8) | VMCODEBYTE(i->pc++);
            if (i->tos)
                i->pc += tmp;
            else
                i->tos = Pop(i);
            break;
        case OP_BRF:
            for (tmp = 0, cnt = sizeof(VMUVALUE); --cnt >= 0; )
                tmp = (tmp << 8) | VMCODEBYTE(i->pc++);
            if (!i->tos)
                i->pc += tmp;
            i->tos = Pop(i);
            break;
        case OP_BRFSC:
            for (tmp = 0, cnt = sizeof(VMUVALUE); --cnt >= 0; )
                tmp = (tmp << 8) | VMCODEBYTE(i->pc++);
            if (!i->tos)
                i->pc += tmp;
            else
                i->tos = Pop(i);
            break;
        case OP_BR:
            for (tmp = 0, cnt = sizeof(VMUVALUE); --cnt >= 0; )
                tmp = (tmp << 8) | VMCODEBYTE(i->pc++);
            i->pc += tmp;
            break;
        case OP_NOT:
            i->tos = (i->tos ? VMFALSE : VMTRUE);
            break;
        case OP_NEG:
            i->tos = -i->tos;
            break;
        case OP_ADD:
            tmp = Pop(i);
            i->tos = tmp + i->tos;
            break;
        case OP_SUB:
            tmp = Pop(i);
            i->tos = tmp - i->tos;
            break;
        case OP_MUL:
            tmp = Pop(i);
            i->tos = tmp * i->tos;
            break;
        case OP_DIV:
            tmp = Pop(i);
            i->tos = (i->tos == 0 ? 0 : tmp / i->tos);
            break;
        case OP_REM:
            tmp = Pop(i);
            i->tos = (i->tos == 0 ? 0 : tmp % i->tos);
            break;
        case OP_BNOT:
            i->tos = ~i->tos;
            break;
        case OP_BAND:
            tmp = Pop(i);
            i->tos = tmp & i->tos;
            break;
        case OP_BOR:
            tmp = Pop(i);
            i->tos = tmp | i->tos;
            break;
        case OP_BXOR:
            tmp = Pop(i);
            i->tos = tmp ^ i->tos;
            break;
        case OP_SHL:
            tmp = Pop(i);
            i->tos = tmp << i->tos;
            break;
        case OP_SHR:
            tmp = Pop(i);
            i->tos = tmp >> i->tos;
            break;
        case OP_LT:
            tmp = Pop(i);
            i->tos = (tmp < i->tos ? VMTRUE : VMFALSE);
            break;
        case OP_LE:
            tmp = Pop(i);
            i->tos = (tmp <= i->tos ? VMTRUE : VMFALSE);
            break;
        case OP_EQ:
            tmp = Pop(i);
            i->tos = (tmp == i->tos ? VMTRUE : VMFALSE);
            break;
        case OP_NE:
            tmp = Pop(i);
            i->tos = (tmp != i->tos ? VMTRUE : VMFALSE);
            break;
        case OP_GE:
            tmp = Pop(i);
            i->tos = (tmp >= i->tos ? VMTRUE : VMFALSE);
            break;
        case OP_GT:
            tmp = Pop(i);
            i->tos = (tmp > i->tos ? VMTRUE : VMFALSE);
            break;
        case OP_LIT:
            for (tmp = 0, cnt = sizeof(VMUVALUE); --cnt >= 0; )
                tmp = (tmp << 8) | VMCODEBYTE(i->pc++);
            CPush(i, i->tos);
            i->tos = tmp;
            break;
        case OP_GREF:
            for (tmp = 0, cnt = sizeof(VMUVALUE); --cnt >= 0; )
                tmp = (tmp << 8) | VMCODEBYTE(i->pc++);
            CPush(i, i->tos);
            i->tos = i->image->variables[tmp];
            break;
        case OP_GSET:
            for (tmp = 0, cnt = sizeof(VMUVALUE); --cnt >= 0; )
                tmp = (tmp << 8) | VMCODEBYTE(i->pc++);
            i->image->variables[tmp] = i->tos;
            i->tos = Pop(i);
            break;
        case OP_LREF:
            tmpb = (int8_t)VMCODEBYTE(i->pc++);
            CPush(i, i->tos);
            i->tos = i->fp[(int)tmpb];
            break;
        case OP_LSET:
            tmpb = (int8_t)VMCODEBYTE(i->pc++);
            i->fp[(int)tmpb] = i->tos;
            i->tos = Pop(i);
            break;
        case OP_VREF:
            tmp = Pop(i);
            ind = i->tos;
            CheckObjNumber(i, tmp);
            hdr = GetObjHdr(i->image, tmp);
            CheckObjType(i, hdr, PROTO_VECTOR);
            if (ind < 0 || ind >= GetVectorSize(hdr))
                Abort(i, "array subscript out of bounds: %d", ind);
            i->tos = GetVectorBase(hdr)[ind];
            break;
        case OP_VSET:
            ind = Pop(i);
            tmp = Pop(i);
            CheckObjNumber(i, tmp);
            hdr = GetObjHdr(i->image, tmp);
            CheckObjType(i, hdr, PROTO_VECTOR);
            if (ind < 0 || ind >= GetVectorSize(hdr))
                Abort(i, "array subscript out of bounds: %d", ind);
            GetVectorBase(hdr)[ind] = i->tos;
            i->tos = Pop(i);
            break;
        case OP_RESERVE:
            tmp = VMCODEBYTE(i->pc++);
            Reserve(i, tmp);
            break;
        case OP_CALL:
            i->argc = VMCODEBYTE(i->pc++);
            StartCode(i, i->tos);
            break;
        case OP_RETURN:
            i->code = i->fp[F_CODE];
            i->cbase = GetBVectorBase(GetObjHdr(i->image, i->code));
            i->pc = i->cbase + i->fp[F_PC];
            i->sp = i->fp;
            Drop(i, VMCODEBYTE(i->pc - 1));
            i->fp = i->stack + i->fp[F_FP];
            break;
        case OP_DROP:
            i->tos = Pop(i);
            break;
        case OP_GETREG:
            ind = VMCODEBYTE(i->pc++);
            CPush(i, i->tos);
            i->tos = VM_getreg(ind);
            break;
        case OP_SETREG:
            ind = VMCODEBYTE(i->pc++);
            VM_setreg(ind, i->tos);
            i->tos = Pop(i);
            break;
        case OP_STRING:
            DoStringOp(i, VMCODEBYTE(i->pc++));
            break;
        case OP_PRINT:
            DoPrint(i, VMCODEBYTE(i->pc++));
            break;
        default:
            Abort(i, "undefined opcode 0x%02x", VMCODEBYTE(i->pc - 1));
            break;
        }
    }
}

static void StartCode(Interpreter *i, VMVALUE object)
{
    if (object & INTRINSIC_FLAG) {
        VMUVALUE index = object & ~INTRINSIC_FLAG;
        if (index < IntrinsicCount)
            (*VMINTRINSIC(index))(i);
        else
            Abort(i, "not code object: %d", object);
    }
    else {
        ObjectHdr *hdr;
        CheckObjNumber(i, object);
        hdr = GetObjHdr(i->image, object);
        if (GetObjPrototype(hdr) == PROTO_CODE) {
            VMVALUE tmp = (VMVALUE)(i->fp - i->stack);
            i->fp = i->sp;
            Reserve(i, F_SIZE);
            i->fp[F_CODE] = i->code;
            i->fp[F_FP] = tmp;
            i->fp[F_PC] = (VMVALUE)(i->pc - i->cbase);
            i->code = object;
            i->cbase = i->pc = GetBVectorBase(hdr);
        }
        else
            Abort(i, "not code object: %d", object);
    }
}

static void DoStringOp(Interpreter *i, int op)
{
    switch (op) {
    case OP_ADD:
        i->tos = StringCat(i);
        break;
    case OP_LT:
        i->tos = (StringCmp(i) < 0 ? VMTRUE : VMFALSE);
        break;
    case OP_LE:
        i->tos = (StringCmp(i) <= 0 ? VMTRUE : VMFALSE);
        break;
    case OP_EQ:
        i->tos = (StringCmp(i) == 0 ? VMTRUE : VMFALSE);
        break;
    case OP_NE:
        i->tos = (StringCmp(i) != 0 ? VMTRUE : VMFALSE);
        break;
    case OP_GE:
        i->tos = (StringCmp(i) >= 0 ? VMTRUE : VMFALSE);
        break;
    case OP_GT:
        i->tos = (StringCmp(i) > 0 ? VMTRUE : VMFALSE);
        break;
    default:
        Abort(i, "undefined string opcode 0x%02x", op);
        break;
    }
}

static VMVALUE StringCat(Interpreter *i)
{
    VMVALUE hStr1 = *i->sp;
    VMVALUE hStr2 = i->tos;
    uint8_t *str1, *str2, *val;
    VMUVALUE len1, len2;
    ObjectHdr *hdr;
    VMVALUE hVal;
    
    /* get the first string pointer and length */
    CheckObjNumber(i, hStr1);
    hdr = GetObjHdr(i->image, hStr1);
    CheckObjType(i, hdr, PROTO_BVECTOR);
    len1 = (VMVALUE)GetVectorSize(hdr);

    /* get the second string pointer and length */
    CheckObjNumber(i, hStr2);
    hdr = GetObjHdr(i->image, hStr2);
    CheckObjType(i, hdr, PROTO_BVECTOR);
    len2 = (VMVALUE)GetVectorSize(hdr);
    
    /* allocate new string and copy both source strings to it */
    hVal = StringAlloc(i->image, len1 + len2);
    val = GetBVectorBase(GetObjHdr(i->image, hVal));
    str1 = GetBVectorBase(GetObjHdr(i->image, hStr1));
    str2 = GetBVectorBase(GetObjHdr(i->image, hStr2));
    memcpy(val, str1, len1);
    memcpy(val + len1, str2, len2);
    Drop(i, 1);
    
    /* return the new string */
    return hVal;
}

static int StringCmp(Interpreter *i)
{
    VMVALUE hStr1 = *i->sp;
    VMVALUE hStr2 = i->tos;
    VMUVALUE cnt, cmp = 0;
    uint8_t *str1, *str2;
    VMUVALUE len1, len2;
    ObjectHdr *hdr;
    
    /* get the first string pointer and length */
    CheckObjNumber(i, hStr1);
    hdr = GetObjHdr(i->image, hStr1);
    CheckObjType(i, hdr, PROTO_BVECTOR);
    str1 = GetBVectorBase(hdr);
    len1 = (VMVALUE)GetVectorSize(hdr);

    /* get the second string pointer and length */
    CheckObjNumber(i, hStr2);
    hdr = GetObjHdr(i->image, hStr2);
    CheckObjType(i, hdr, PROTO_BVECTOR);
    str2 = GetBVectorBase(hdr);
    len2 = (VMVALUE)GetVectorSize(hdr);
    
    /* handle the case where the first string is shorter than the second */
    if (len1 < len2) {
        for (cnt = len1; cnt > 0; --cnt) {
            if ((cmp = VMCODEBYTE(str1++) - VMCODEBYTE(str2++)) != 0)
                break;
        }
        if (cnt <= 0)
            cmp = -1;
    }
    
    /* handle the case where the first string is longer or the same length as the first */
    else {
        for (cnt = len2; cnt > 0; --cnt) {
            if ((cmp = VMCODEBYTE(str1++) - VMCODEBYTE(str2++)) != 0)
                break;
        }
        if (cnt <= 0)
            cmp = (len1 == len2 ? 0 : 1);
    }

    /* return the comparision result */
    return cmp;
}

static void DoPrint(Interpreter *i, int op)
{
    switch (op) {
    case PRINT_INT:
        PrintInteger(i, i->tos);
        i->tos = Pop(i);
        break;
    case PRINT_STR:
        PrintString(i, i->tos);
        i->tos = Pop(i);
        break;
    case PRINT_TAB:
        PrintTab(i);
        break;
    case PRINT_NL:
        PrintC(i, '\n');
        break;
    default:
        Abort(i, "undefined print opcode 0x%02x", op);
        break;
    }
}

static void PrintInteger(Interpreter *i, VMVALUE value)
{
    char buf[16], *p;
    sprintf(buf, "%ld", (long)value);
    for (p = buf; *p != '\0'; )
        PrintC(i, *p++);
}

static void PrintString(Interpreter *i, VMVALUE value)
{
    ObjectHdr *hdr;
    uint8_t *str;
    int len;
    CheckObjNumber(i, value);
    hdr = GetObjHdr(i->image, value);
    CheckObjType(i, hdr, PROTO_BVECTOR);
    str = GetBVectorBase(hdr);
    len = (VMVALUE)GetVectorSize(hdr);
    while (--len >= 0)
        PrintC(i, VMCODEBYTE(str++));
}

static void PrintTab(Interpreter *i)
{
    do {
        PrintC(i, ' ');
    } while ((i->linePos % 8) != 0);
}

static void PrintC(Interpreter *i, int ch)
{
    VM_putchar(ch);
    if (ch == '\n')
        i->linePos = 0;
    else
        ++i->linePos;
}

void ShowStack(Interpreter *i)
{
    VMVALUE *p;
    for (p = i->stack; p <= i->sp; ++p) {
        if (p == i->fp)
            VM_printf(" <fp>");
        VM_printf(" %d", *p);
    }
    VM_printf(" %d\n", i->tos);
}


void StackOverflow(Interpreter *i)
{
    Abort(i, "stack overflow");
}

void Warn(const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    VM_printf("warning: ");
    VM_vprintf(fmt, ap);
    VM_putchar('\n');
    va_end(ap);
}

void Abort(Interpreter *i, const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    VM_printf("abort: ");
    VM_vprintf(fmt, ap);
    VM_putchar('\n');
    va_end(ap);
    if (i)
        longjmp(i->errorTarget, 1);
    else
        exit(1);
}
